##################################################################################
#  SENSIML CONFIDENTIAL                                                          #
#                                                                                #
#  Copyright (c) 2017  SensiML Corporation.                                      #
#                                                                                #
#  The source code contained or  described  herein and all documents related     #
#  to the  source  code ("Material")  are  owned by SensiML Corporation or its   #
#  suppliers or licensors. Title to the Material remains with SensiML Corpora-   #
#  tion  or  its  suppliers  and  licensors. The Material may contain trade      #
#  secrets and proprietary and confidential information of SensiML Corporation   #
#  and its suppliers and licensors, and is protected by worldwide copyright      #
#  and trade secret laws and treaty provisions. No part of the Material may      #
#  be used,  copied,  reproduced,  modified,  published,  uploaded,  posted,     #
#  transmitted, distributed,  or disclosed in any way without SensiML's prior    #
#  express written permission.                                                   #
#                                                                                #
#  No license under any patent, copyright,trade secret or other intellectual     #
#  property  right  is  granted  to  or  conferred upon you by disclosure or     #
#  delivery of the Materials, either expressly, by implication,  inducement,     #
#  estoppel or otherwise.Any license under such intellectual property rights     #
#  must be express and approved by SensiML in writing.                           #
#                                                                                #
#  Unless otherwise agreed by SensiML in writing, you may not remove or alter    #
#  this notice or any other notice embedded in Materials by SensiML or SensiML's #
#  suppliers or licensors in any way.                                            #
#                                                                                #
##################################################################################
from __future__ import print_function
import os
import sys
import inspect
import json
from logging import debug
import pandas as pd
from pandas import DataFrame
from copy import deepcopy
from numbers import Number
import datetime
from sensiml.render import BaseRenderer

try:
    from urllib.parse import urlencode
except ImportError:
    from urllib import urlencode


import sensiml.base.utility as utility
from sensiml.datamanager.pipeline import PipelineStep, PipelineError
from sensiml.datamanager.knowledgepack import KnowledgePack
from sensiml.method_calls.functioncalls import FunctionCalls
from sensiml.method_calls.selectorcallset import SelectorCallSet
from sensiml.method_calls.generatorcallset import GeneratorCallSet
from sensiml.method_calls.trainandvalidationcall import TrainAndValidationCall
from sensiml.datamanager.modelresults import ModelResultSet
from sensiml.datamanager.deviceconfig import DeviceConfig


class SandboxPipeline():
    """Base class for a sandbox pipeline consisting of steps"""

    def __init__(self, steps=[]):
        self._steps = steps

    def add(self, item):
        """Appends a PipelineStep object to the pipeline while performing integrity checks.

            Args:
                item: a QueryCall, FunctionCall, GeneratorSet, SelectorSet, or TrainAndValidationCall object

            Returns:
                nothing, or raises an exception if integrity check fails
        """
        assert isinstance(item, PipelineStep)
        self._steps.append(item)

        try:
            self._check_integrity()
        except:
            self._steps.pop()
            raise

    def replace(self, item, index):
        """Replace a pipeline step that has the same name.

        Args:
            item: a QueryCall, FunctionCall, GeneratorSet, SelectorSet, or TrainAndValidationCall object
            index (int): index of pipeline step to replace

            Returns:
                nothing, or raises an exception if integrity check fails
        """
        assert isinstance(item, PipelineStep)
        assert (item.name == self._get_step_name(index))
        backup_item = deepcopy(self._steps[index])

        self._steps[index] = item

        try:
            self._check_integrity()
        except:
            self._steps[index] = backup_item
            raise

    def linear_pipeline(self, item):
        """A linear step will automatically set the previous step's output to the currents step's input.

            Args:
                item (pipeline step): a pipeline step to modify

            Returns:
                pipeline step: pipeline step with updated input_data and outputs attributes
        """

        # if this is the first step
        if len(self._steps) == 0:
            item.outputs = ['temp.raw']
        else:
            item.input_data = self._steps[-1].outputs[0]
            # Count the number of times the name has been used to get unique
            # input/output names
            index = len([
                1 for s in self._steps
                if 'temp.{}'.format(
                    item.name.replace(' ', '_')) in s.outputs[0]
            ])
            item.outputs = [
                'temp.{}{}'.format(item.name.replace(' ', '_'), index)
            ]

            if item._to_dict()['type'] == 'generatorset':
                item.outputs.append('temp.features.{}{}'.format(
                    item.name.replace(' ', '_'), index))
            if len(self._steps[-1].outputs) > 1:
                item.feature_table = self._steps[-1].outputs[-1]
                item.outputs.append('temp.features.{}{}'.format(
                    item.name.replace(' ', '_'), index))

        return item

    def replace_linear_pipeline(self, item, index):
        item.input_data = self._steps[index].input_data
        item.outputs = self._steps[index].outputs
        if len(self._steps[index].outputs) > 1:
            item.feature_table = self._steps[index].feature_table

        return item

    def remove(self, item):
        """Removes the item from the pipeline.

            Args:
                item (pipeline step)
        """
        if item in self._steps:
            self._steps.remove(item)

    def clear(self):
        """Clears all items from the pipeline."""
        self._steps = []

    def to_list(self):
        """Returns a list representation of the pipeline."""
        a = []
        if self._steps:
            for step in self._steps:
                if isinstance(step, FunctionCalls):
                    for o in step._to_list():
                        a.append(o)
                else:
                    if (isinstance(step, dict)):
                        a.append(step)
                    else:
                        a.append(step._to_dict())
        return a

    def describe(self, show_params=True, show_set_params=False):
        """Prints a formatted description of the pipeline"""
        describe_pipeline(self.to_list(), show_params=show_params,
                          show_set_params=show_set_params)

    def _get_pipeline_step_value(self, step, key='type'):
        """Returns the value associated with the specified key from the specified step index.

            Args:
                step (int): step index
                key (str): key to get the value for

            Returns:
                if it exists, the contents associated with the key, otherwise None
        """
        if len(self._steps) == 0:
            print('There are no steps associated with this pipeline.')
            return None

        if len(self._steps) < step:
            print('This pipeline step does not exist.')
            return None

        if isinstance(self._steps[step], dict):
            return self._steps[step].get(key, None)

        return self._steps[step]._to_dict().get(key, None)

    def _get_step_name(self, index):
        if isinstance(self._steps[index], dict):
            return self._steps[index]['name']
        else:
            return self._steps[index].name

    def _check_duplicate_outputs(self, item):
        for i, step in enumerate(self._steps):
            if item.name == self._get_step_name(i):
                return i

        return None

    def _check_integrity(self):
        """Checks the integrity of the pipeline as a whole.

            Returns:
                 True if successful, otherwise raises an exception
        """
        _aggregated_errors = []

        def _aggregate_errors(fmt_string, errors=[], *args):
            # Useful aggregation of error strings
            for error in errors:
                err_string = fmt_string.format(error, *args)
                _aggregated_errors.append(err_string)
                # _aggregated_errors.append(fmt_string.format(args, error))
            return

        output_vars = []
        step_num = 1
        for step in self._steps:
            is_last_step = True if step_num == len(self._steps) else False
            if isinstance(step, PipelineStep):
                errors = step._check_inputs(output_vars)
                if errors:
                    _aggregate_errors(
                        '[{1}] Step "{2}" -- Unmatched input: {0}', errors,
                        step_num, step.name)
                errors = step._check_outputs(output_vars, is_last_step)
                if errors:
                    _aggregate_errors(
                        '[{1}] Step "{2}" -- Duplicate output: {0}', errors,
                        step_num, step.name)
                output_vars.extend(step.outputs)
            else:
                raise PipelineError('Item is not a valid pipeline step.')
            step_num += 1

            # Ensure that the set objects contain the minimum number of call
            # objects
            missing_call_string = "The {0} set does not contain any {1} calls. Please add one or more {1} calls before adding the step to the sandbox."
            if isinstance(step,
                          GeneratorCallSet) and not len(step.generator_calls):
                raise PipelineError(
                    missing_call_string.format('feature generator',
                                               'generator'))
            elif isinstance(step, SelectorCallSet) and not len(step.selectors):
                raise PipelineError(
                    missing_call_string.format('feature selector', 'selector'))
            elif isinstance(step, TrainAndValidationCall):
                if not len(step.classifiers):
                    raise PipelineError(
                        missing_call_string.format('train and validation',
                                                   'classifier'))
                if not len(step.optimizers):
                    raise PipelineError(
                        missing_call_string.format('train and validation',
                                                   'optimizer'))
                if not len(step.validation_methods):
                    raise PipelineError(
                        missing_call_string.format('train and validation',
                                                   'validation method'))

        if not _aggregated_errors:
            return True
        else:
            raise PipelineError('Pipeline failed integrity check: ' +
                                '\n'.join(_aggregated_errors))

    def __str__(self):
        pipeline_string = ''
        pipeline_list = []
        for step in self._steps:
            if isinstance(step, FunctionCalls):
                for o in step._to_list():
                    pipeline_list.append(o)
            else:
                if isinstance(step, dict):
                    pipeline_list.append(step)
                else:
                    pipeline_list.append(step._to_dict())
        for step in pipeline_list:
            if 'name' in step:
                pipeline_string += '{name} ({type})\n'.format(**step)
            else:
                pipeline_string += '{type}\n'.format(**step)
            if 'inputs' in step:
                pipeline_string += '    Inputs: \n'
                for input_ in step['inputs']:
                    pipeline_string += '        {0}\n'.format(input_)
            if 'outputs' in step:
                pipeline_string += '    Outputs: \n'
                for output_ in step['outputs']:
                    pipeline_string += '        {0}\n'.format(output_)
            pipeline_string += '\n'

        return pipeline_string


class Sandbox(object):
    """Base class for a sandbox."""

    def __init__(self, connection, project):
        """Initializes a sandbox instance."""
        self._uuid = ''
        self._name = ''
        self._dirty = False
        self._connection = connection
        self._project = project
        self._pipeline = SandboxPipeline()
        self._cache_enabled = True
        self._results = ModelResultSet(project, self)
        self._execution_summary = None
        self._features = None
        self._device_config = DeviceConfig()
        self._created_at = None

    @property
    def uuid(self):
        return self._uuid

    @uuid.setter
    def uuid(self, value):
        self._uuid = value

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._dirty = True
        self._name = value

    @property
    def pipeline(self):
        return self._pipeline

    @property
    def cache(self):
        return self._cache_enabled

    @property
    def results(self):
        return self._results

    @property
    def device_config(self):
        return self._device_config

    @property
    def created_at(self):
        """Date of the Pipeline creation"""
        return self._created_at

    @created_at.setter
    def created_at(self, value):
        if value:
            self._created_at = datetime.datetime.strptime(
                value[:-6], '%Y-%m-%dT%H:%M:%S.%f')

    @property
    def cost_preview(self):
        print('The cost_preview is no longer available. To view device costs look at a KnowledgePack cost_report.\n')

    def _check_sandbox_inserted(self, insert=False):
        """ Checks for a uuid and warns the user if none exists """
        if self.uuid == '':
            if insert:
                print('Sandbox was not on server, attempting to insert now...', end='')
                self.insert()
                print('Success')
            else:
                raise Exception("Sandbox does not exist on the server")
        return True

    def _init_device_config(self, data):
        self._device_config = DeviceConfig(data.get('device_config', None))

    def insert(self):
        """Calls the REST API to insert a new object on to the server."""
        url = 'project/{0}/sandbox/'.format(self._project.uuid)
        sandbox_info = {
            'name': self.name,
            'pipeline': self.pipeline.to_list(),
            'cache_enabled': self.cache,
            'device_config': self.device_config
        }
        response = self._connection.request('post', url, sandbox_info)
        response_data, err = utility.check_server_response(response)
        if err is False:
            self.uuid = response_data['uuid']
            self._init_device_config(response_data)
            self._dirty = False

    def update(self):
        """Calls the REST API to update the object on the server."""
        self._pipeline._check_integrity()
        if self._check_sandbox_inserted():
            url = 'project/{0}/sandbox/{1}/'.format(self._project.uuid,
                                                    self.uuid)
            sandbox_info = {
                'name': self.name,
                'pipeline': self.pipeline.to_list(),
                'cache_enabled': self.cache,
                'device_config': self.device_config
            }
            response = self._connection.request('put', url, sandbox_info)
            response_data, err = utility.check_server_response(response)
            if err is False:
                self._init_device_config(response_data)
                self._dirty = False

    def delete(self):
        """Calls the REST API to delete the object from the server."""
        url = 'project/{0}/sandbox/{1}/'.format(self._project.uuid, self.uuid)
        if self._check_sandbox_inserted(insert=False):
            response = self._connection.request(
                'delete',
                url, )
            response_data, err = utility.check_server_response(response)
            self._dirty = False
            self.uuid = ''

    def knowledgepack(self, kp_uuid):
        """Gets the KnowledgePack(s) created by the sandbox.

            Returns:
                a KnowledgePack instance, list of instances, or None
        """
        url = 'project/{0}/sandbox/{1}/knowledgepack/{2}/'.format(
            self._project.uuid, self.uuid, kp_uuid)
        response = self._connection.request(
            'get',
            url, )
        response_data, err = utility.check_server_response(response)
        if err is False:
            kp = KnowledgePack(self._connection, self._project.uuid, self.uuid)
            kp.initialize_from_dict(response_data)
            return kp

    def get_knowledgepacks(self):
        """Gets the KnowledgePack(s) created by the sandbox.

            Returns:
                a KnowledgePack instance, list of instances, or None
        """
        url = 'project/{0}/sandbox/{1}/knowledgepack/'.format(
            self._project.uuid, self.uuid)
        response = self._connection.request(
            'get',
            url, )
        response_data, err = utility.check_server_response(response)
        if err is False:
            return pd.DataFrame(response_data)

    def _handle_result(self, data, pipeline_step=-1):
        """Decides what type of results have been returned and creates a DataFrame or stores model results accordingly."""

        if data.get('status'):
            print(data.get('status'))
            return None, None

        execution_type = data.get('execution_type')

        if data.get('number_of_pages', 0) > 1:
            print('\nRetrieving page {0} of {1}.'.format(
                data.get('page_index', 1), data['number_of_pages']))
            if data['number_of_pages'] > 1:
                print("To get more pages, use: \n" +
                      "next_results, next_stats = dsk.pipeline.get_results(page_index=<desired page number>).")

        summary = pd.DataFrame(data.get('summary', [])).fillna('')
        extra = pd.DataFrame(data.get('extra', []))
        results = data.get('results', [])

        self._execution_summary = summary
        self._features = extra

        if (self.pipeline._get_pipeline_step_value(pipeline_step,
                                                   key='type') == 'tvo' and
                execution_type == 'pipeline') or execution_type == 'auto':
            if isinstance(results, list):
                data = []
                for result in results:
                    temp_results = ModelResultSet(
                        deepcopy(self._project), deepcopy(self))
                    temp_results.initialize_from_dict(result)
                    data.append(temp_results)
            else:
                self._results = ModelResultSet(self._project, self)
                self._results.initialize_from_dict(results)
                data = self._results

        else:
            if isinstance(results, list):
                data = map(lambda x: DataFrame(x), results)
            else:
                data = DataFrame(results)

        if execution_type == 'auto' and 'fitness' in extra.columns:
            return data, {
                'execution_summary': summary,
                'fitness_summary': extra
            }
        else:
            return data, {'execution_summary': summary, 'features': extra}

    def intermediate_data(self, pipeline_step, page_index=0):
        """Retrieves intermediate pipeline step data from a previously executed pipeline.

        Args:
            pipeline_step (int): The pipeline step to retrieve from and executed pipeline.
            page_index (int): the index to pull form the cache

        Returns:
            (DataFrame or ModelResultSet): result of executed pipeline, specified by the sandbox.

        """
        url = 'project/{0}/sandbox/{1}/data/'.format(self._project.uuid,
                                                     self.uuid)
        response = self._connection.request(
            'get', url, data={'pipeline_step': pipeline_step,
                              'page_index': page_index})

        data, err = utility.check_server_response(response)
        if err is False:
            response_data = data
            data, summary = self._handle_result(
                response_data, pipeline_step=pipeline_step)

            return data, summary

    def get_metrics_set(self, data):
        """retrieves metrics set from server

        Args:
            data (dict): dictionary containing
                        - 'y_true' and 'y_pred' keys
                        - y_true and y_pred values must be lists

        Returns:
            (dict): dictionary containing metrics result set from server"""
        url = 'get_metrics_set/'
        response = self._connection.request('get', url, data)
        data, err = utility.check_server_response(response)
        if err is False:
            response_data = data

            return response_data

    def execute(self):
        """Executes the sandbox pipeline and returns results.

            Returns:
                (DataFrame or ModelResultSet): result of executed pipeline, specified by the sandbox
                (dict): execution summary including execution time and whether cache was used for each
                step; also contains a feature cost table if applicable
        """
        self.submit()

        return utility.wait_for_pipeline_result(self,
                                                lock=True,
                                                wait_time=1,
                                                silent=True,
                                                page_index=1,
                                                )

    def async_submit(self, data):
        """Submits the sandbox for asynchronous execution and returns the task status."""
        if self._dirty:
            self.update()

        url = 'project/{0}/sandbox-async/{1}/'.format(self._project.uuid,
                                                      self.uuid)
        response = self._connection.request('post', url, data)
        response_data, err = utility.check_server_response(response)
        if err is False:
            return response_data

    def auto(self, auto_params, seed, renderer=None):
        """Submits the automation job for asynchronous execution and returns the task status."""
        # If the sandbox contains anything except a querycall or featurefilecall, and it
        # is not a custom seed, do not execute the request
        if renderer is None:
            renderer = BaseRenderer()
        if seed.name == 'Custom':
            if self.pipeline.to_list()[-1]['type'] != 'tvo':
                renderer.render(
                    "Only custom pipelines that end with a tvo call can be automated.")
                return 0
            renderer.render("Running Auto Pipeline with Custom Seed:\n")
            describe_pipeline(self.pipeline.to_list())
        elif len(self.pipeline.to_list()) == 0:
            renderer.render(
                "Running Auto Pipeline with {} Seed and Auto Query:\n".format(seed.name))
        elif len(self.pipeline.to_list()) == 1 and self.pipeline.to_list()[0]['type'] in ['query', 'featurefile']:
            renderer.render(
                "Running Auto Pipeline with {} Seed:\n".format(seed.name))
            full_pipeline = deepcopy(self.pipeline.to_list()) + seed.pipeline
        elif len(self.pipeline.to_list()) > 1 and self.pipeline.to_list()[0]['type'] in ['query']:
            renderer.render(
                "Running Auto Pipeline with {} Seed:\n".format(seed.name))
            full_pipeline = deepcopy(self.pipeline.to_list()) + seed.pipeline
        elif len(self.pipeline.to_list()) <= 2 and 'segmenter' in [step['type'] for step in self.pipeline.to_list()] and \
                set([step['type'] for step in self.pipeline.to_list()]).issubset(['query', 'featurefile', 'segmenter']):
            segmenter_name = [step['name'] for step in self.pipeline.to_list(
            ) if step['type'] == 'segmenter'][0]
            renderer.render("Running Auto Pipeline with {} Seed and {} Segmenter:\n".format(
                seed.name, segmenter_name))
            full_pipeline = deepcopy(self.pipeline.to_list()) + seed.pipeline
        else:
            renderer.render("This pipeline contains steps that could be overwritten by the {}\n".format(seed.name) +
                            "seed. Either run this pipeline with the Custom seed or remove steps from the \n" +
                            "pipeline (query, data file, and segmenter steps are okay).")
            return 0

        data = {
            'run_parallel': True,
            'execution_type': 'auto',
            'auto_params': auto_params
        }

        self.async_submit(data)
        return 1

    def retrieve_auto(self):
        result = self.retrieve(execution_type='auto')
        return result

    def autosegment_search(self, params, run_parallel=True):
        """Submits the grid search for asynchronous execution and returns the task status."""
        data = {
            'run_parallel': run_parallel,
            'execution_type': 'autosegment_search',
            'seg_params': params
        }

        return self.async_submit(data)

    def retrieve_autosegment(self):
        result = self.retrieve()
        return result

    def retrieve(self, silent=False, page_index=0, **kwargs):
        """Gets the result of a prior asynchronous execution of the sandbox.

            Returns:
                (DataFrame or ModelResultSet): result of executed pipeline, specified by the sandbox
                (dict): execution summary including execution time and whether cache was used for each
                step; also contains a feature cost table if applicable
        """
        if self._dirty:
            self.update()

        url = 'project/{0}/sandbox-async/{1}/?{2}'.format(
            self._project.uuid, self.uuid,
            urlencode({
                'page_index': page_index
            }))
        response = self._connection.request('get', url)
        data, err = utility.check_server_response(response)
        if err is False:
            response_data = data or {}
            return utility.check_pipeline_status(
                response_data,
                self._handle_result,
                silent=silent,
            )

    def kill_pipeline(self):
        url = 'project/{0}/sandbox-async/{1}/'.format(self._project.uuid,
                                                      self.uuid)
        response = self._connection.request('delete', url)
        response_data, err = utility.check_server_response(response)
        if err is False:
            print(response_data)
            return response_data

    def submit(self):
        """Submits a pipeline for asynchronous execution."""
        data = {'execution_type': 'pipeline'}
        return self.async_submit(data)

    def grid_search(self, grid_params, run_parallel=False):
        """Submits the grid search for asynchronous execution and returns the task status."""
        data = {
            'run_parallel': run_parallel,
            'execution_type': 'grid_search',
            'grid_params': grid_params
        }

        return self.async_submit(data)

    def retrieve_gridsearch(self):
        result = self.retrieve()
        return result

    def refresh(self):
        """Calls the REST API and populates the local sandbox from the server."""
        url = 'project/{0}/sandbox/{1}/'.format(self._project.uuid, self.uuid)
        response = self._connection.request('get', url)
        data, err = utility.check_server_response(response)
        if err is False:
            self.uuid = data['uuid']
            self.name = data['name']
            self.created = data.get('created_at', None)
            if not data.get('cache_enabled', True):
                self.disable_cache()
            self._init_device_config(data)
            self.pipeline._steps = data['pipeline']
            self._dirty = False

    def delete_cache(self):
        """Calls the REST API and deletes the server cache for the sandbox."""
        if self._check_sandbox_inserted(insert=False):
            url = 'project/{0}/sandbox/{1}/cache/'.format(
                self._project.uuid, self.uuid)
            response = self._connection.request('DELETE', url)
            data, err = utility.check_server_response(response)
            if err is False:
                print(data)

    def initialize_from_dict(self, init_dict):
        """Populates a single sandbox from a dictionary of properties.

            Args:
                (dict): contains uuid, name, cache_enabled, and pipeline
        """
        self.uuid = init_dict['uuid']
        self.name = init_dict['name']
        self.created_at = init_dict.get('created_at', None)

        if not init_dict.get('cache_enabled', True):
            self.disable_cache()

        if len(init_dict['pipeline']) != 0:
            self._pipeline._steps = init_dict['pipeline']
        else:
            self._pipeline._steps = []
            self._init_device_config(init_dict)
        self._dirty = False

    def add_step(self, step):
        """Adds the step to the sandbox pipeline.

            Args:
                step (pipeline step): a QueryCall, FunctionCall, GeneratorSet, SelectorSet, or
                                      TrainAndValidationCall object
        """
        self._pipeline.add(step)
        self._dirty = True

    def update_step(self, step, index):
        """Update a step with new values

            Args:
                step (pipeline step): a QueryCall, FunctionCall, GeneratorSet, SelectorSet, or
                                      TrainAndValidationCall object
                index (int): index of the pipeline step to update

        """
        step = self._pipeline.replace_linear_pipeline(step, index)
        self._pipeline.replace(step, index)
        self._dirty = True

    def add_linear_step(self, step, overwrite=True):
        """Adds a linear step to the pipeline (will automatically use the previous step's output as input).

            Args:
                step (pipeline step): a QueryCall, FunctionCall, GeneratorSet, SelectorSet, or
                    TrainAndValidationCall object
                overwrite (boolean): when adding multiple instances of the same transform, set
                    overwrite to False for the additional steps and the first instance will not
                    be overwritten (default is True)

        """
        if not overwrite or self._pipeline._check_duplicate_outputs(
                step) is None:
            step = self._pipeline.linear_pipeline(step)
            self._pipeline.add(step)
            self._dirty = True
        else:
            self.update_step(step,
                             self._pipeline._check_duplicate_outputs(step))

    def remove_step(self, step):
        """Removes a step from the pipeline.

            Args:
                step (pipeline step): a QueryCall, FunctionCall, GeneratorSet, SelectorSet, or
                                      TrainAndValidationCall object to remove from the pipeline
        """
        self._pipeline.remove(step)
        self._dirty = True

    def clear(self):
        """Clears the sandbox of all steps."""
        self._pipeline.clear()
        self._dirty = True

    def enable_cache(self):
        """Enables the use of caching for the sandbox on the server."""
        self._cache_enabled = True
        self._dirty = True

    def disable_cache(self):
        """Disables the use of caching for the sandbox on the server."""
        self._cache_enabled = False
        self._dirty = True


def describe_pipeline(pipeline, show_params=True, show_set_params=False):
    print("------------------------------------------------------------------------")
    for index, step in enumerate(pipeline):

        if isinstance(step['name'], list):
            name = ', '.join(step['name'])
        else:
            name = step['name']

        print("{0: >2}.     Name: {1: <25} \t\tType: {2: <25}".format(
            index, name, step['type']))
        print("------------------------------------------------------------------------")
        if step.get('set', None) and show_params:
            print_set_items(step, params=show_set_params)
        if step.get('inputs', None) and show_params:
            print(filter_params(step['inputs']))
        if step.get('type') == 'tvo':
            print_tvo(step)
        print("------------------------------------------------------------------------")
    print("")


def filter_params(inputs, num_tabs=1):
    not_params = [
        'input_data', 'group_column',
        'passthrough_columns', 'group_columns', 'remove_columns',
        'label_column', 'feature_table'
    ]
    params = filter(lambda x: x not in not_params, inputs.keys())
    param_values = [
        num_tabs * "\t" + "{0} {1}".format(param + ':',
                                           inputs[param])
        for param in params
    ]
    return '\n'.join(param_values)


def print_set_items(step, params=False):
    for index, set_step in enumerate(step.get('set')):
        print("\t{0: >2}. Name: {1: <25}".format(index,
                                                 set_step['function_name']))
        if params:
            print(filter_params(set_step['inputs'], num_tabs=2))
            print('')


def print_tvo(step):
    map_tvo = {
        'classifiers': 'Classifier:',
        'optimizers': 'Training Algo:',
        'validation_methods': 'Validation Method:'
    }
    for tvo_step in ['classifiers', 'optimizers', 'validation_methods']:
        print("\t{0} {1}".format(map_tvo[tvo_step],
                                 step[tvo_step][0]['name']))
        print(filter_params(step[tvo_step][0]['inputs'], num_tabs=2))
        print('')



##################################################################################
#  SENSIML CONFIDENTIAL                                                          #
#                                                                                #
#  Copyright (c) 2017-18  SensiML Corporation.                                   #
#                                                                                #
#  The source code contained or  described  herein and all documents related     #
#  to the  source  code ("Material")  are  owned by SensiML Corporation or its   #
#  suppliers or licensors. Title to the Material remains with SensiML Corpora-   #
#  tion  or  its  suppliers  and  licensors. The Material may contain trade      #
#  secrets and proprietary and confidential information of SensiML Corporation   #
#  and its suppliers and licensors, and is protected by worldwide copyright      #
#  and trade secret laws and treaty provisions. No part of the Material may      #
#  be used,  copied,  reproduced,  modified,  published,  uploaded,  posted,     #
#  transmitted, distributed,  or disclosed in any way without SensiML's prior    #
#  express written permission.                                                   #
#                                                                                #
#  No license under any patent, copyright,trade secret or other intellectual     #
#  property  right  is  granted  to  or  conferred upon you by disclosure or     #
#  delivery of the Materials, either expressly, by implication,  inducement,     #
#  estoppel or otherwise.Any license under such intellectual property rights     #
#  must be express and approved by SensiML in writing.                           #
#                                                                                #
#  Unless otherwise agreed by SensiML in writing, you may not remove or alter    #
#  this notice or any other notice embedded in Materials by SensiML or SensiML's #
#  suppliers or licensors in any way.                                            #
#                                                                                #
##################################################################################


import os
import IPython
from ipywidgets import widgets
from ipywidgets import Layout, Button, Box, FloatText, Textarea, Dropdown, Label, IntSlider, Checkbox, Text, Button, SelectMultiple
from IPython.display import display
from ipywidgets import IntText
from json import dumps as jdump
from sensiml.widgets.base_widget import BaseWidget

title_item_layout = Layout(
    display='flex',
    flex_flow='row',
    size=16,
    border='solid 2px',
    justify_content='space-between',
    overflow='visible'
)

label_value_layout = Layout(overflow='visible',
                            min_width='150px',)

category_item_layout = Layout(
    display='flex',
    flex_flow='row',
    size=16,
    border='solid 2px',
    justify_content='center',
    background_color='red',
    overflow='visible'
)

form_item_layout = Layout(
    display='flex',
    flex_flow='row',
    overflow='visible',
    justify_content='space-between'
)


pipeline_layout = Layout(display='flex',
                         flex_flow='column',
                         border='solid 2px',
                         align_items='stretch',
                         width='75%',
                         overflow='visible')

GENERATE_INDEX = 0
PLATFORM_INDEX = 1
DEBUG_INDEX = 2
TEST_DATA_INDEX = 3
APPLICATION_INDEX = 4
SAMPLE_RATE_INDEX = 5
DOWNLOAD_TYPE = 6
TEST_DATA = 7
PARENT_INDEX = 8
SUBMODEL_INDEX = 9


def clean_name(name):
    return ''.join(e if e.isalnum() else '_' for e in name)


class HierarchicalWidget(BaseWidget):
    def __init__(self, dsk=None, level='Project', folder='knowledgepacks'):
        self._dsk = dsk
        if not os.path.exists(folder):
            os.makedirs(folder)
        self.setup(level=level)

    def setup(self, level):
        self.kb_description = {'parent': {},
                               'sub': {'Report': 'Report'}}

        self.kb_dict = {'parent': [],
                        'sub': []}

        self.level = level
        self.platforms = self._dsk.platforms()

    def select_platform(self, platform_name):
        self.platform = self.platforms.iloc[platform_name]
        self.platform_name = self.platform['Software Platform']

        if 'curie' in self.platform['Description'].lower():
            self.target_os_widget.options = ['ISPC']
            self.target_os_widget.value = self.target_os_widget.options[0]
            self.app_output_options_widget.options = ['ble', 'led', 'serial']
        elif 'nordic' in self.platform['Description'].lower():
            self.target_os_widget.options = ['NordicSDK']
            self.target_os_widget.value = self.target_os_widget.options[0]
            self.app_output_options_widget.options = ['ble', 'led', 'serial']
            self.app_output_options_widget.value = ['ble']
        elif 'simulator' in self.platform['Description'].lower():
            self.target_os_widget.options = ['x86']
            self.target_os_widget.value = self.target_os_widget.options[0]
            self.app_output_options_widget.options = ['serial']

    def generate_description(self, b):
        parent_name = None
        parent_uuid = None
        description = {}
        if not self.kb_description['parent']:
            print("No Parent Model Set.")
        for key in self.kb_description['parent']:
            description = {clean_name(key): {
                'uuid': self.kb_description['parent'][key], 'results': {}, 'source': 'motion'}}
            parent_name = clean_name(key)
            parent_uuid = self.kb_description['parent'][key]
        for box in self.kb_dict['parent'][0][1:]:
            description[parent_name]['results'].update(
                {format(box.children[0].description.split('-')[0]): clean_name(box.children[0].value)})
        for key in self.kb_description['sub']:
            if key != 'Report':
                sub_description = {'uuid': self.kb_description['sub'][key],
                                   'parent': parent_name,
                                   'segmenter_from': 'parent'}
                description.update({clean_name(key): sub_description})

        kp_uuid = parent_uuid
        board_name = self._dsk.platforms[self.platform_widget.value].board_name

        kp_platform = self._dsk.platforms[self.platform_widget.value].id
        kp_debug = self.debug_widget.value
        kp_test_data = self.test_data_widget.value
        kp_download_type = self.download_type_widget.value
        sample_rate = self.imu_sample_widget.value
        output_options = self.app_output_options_widget.value

        if kp_platform >= 3 and kp_platform < 99:  # Nordic Thingy and not simulator
            if 'ble' not in output_options:
                output_options.insert('ble')  # always output via ble
            kp_application = 'thingy_pme'
        else:
            if 'led' in output_options:
                kp_application = 'LED'
            else:
                kp_application = 'Default'

        if board_name == 'ARM GCC Generic' or board_name == 'x86 GCC Generic':
            kp_application = 'testdata_runner'

        if kp_uuid is not None:
            kp = self._dsk.get_knowledgepack(kp_uuid)
        else:
            #print('Invalid KnowledgePack name.')
            return None

        config = {'target_platform': kp_platform,
                  'test_data': kp_test_data,
                  'debug': kp_debug,
                  'application': kp_application,
                  'sample_rate': sample_rate,
                  'output_options': output_options,
                  'kb_description': description}

        # print(config)

        if kp_download_type == 'Library':
            kp.download_library(config=config, folder='knowledgepacks')
        if kp_download_type == 'Binary':
            kp.download_binary(config=config, folder='knowledgepacks')

    def set_parent_model(self, b):
        index = int(b.icon)

        kp = self._dsk.get_knowledgepack(
            self.kb_items[PARENT_INDEX].children[1].value)
        kp_item = []
        kp_item.append(Box([Label(value=kp.name),
                            Button(description='delete'),
                            ], layout=title_item_layout))

        for key, value in kp.class_map.items():
            kp_item.append(Box([Dropdown(options=self.kb_description['sub'],
                                         description='{} - {}'.format(key, value))],
                               layout=form_item_layout))

        self.kb_description['parent'] = {kp.name: kp.uuid}
        self.kb_dict['parent'] = [kp_item]
        self.generate_pipeline()

    def add_submodel(self, b):
        index = int(b.icon)
        kp = self._dsk.get_knowledgepack(
            self.kb_items[SUBMODEL_INDEX].children[1].value)
        kp_item = []
        if self.kb_description['sub'].get(kp.name, None) is not None:
            return

        if self.kb_description['parent'].get(kp.name, None) is not None:
            return

        kp_item.append(Box([Label(value=kp.name),
                            Button(description='delete'),
                            ], layout=title_item_layout))

        self.kb_description['sub'].update({kp.name: kp.uuid})
        self.kb_dict['sub'].append(kp_item)
        self.generate_pipeline()

    def get_kp_dict(self):
        if self.level.lower() == 'project':
            return {name: value for name, value in self._dsk.project.list_knowledgepacks()[['Name', 'kp_uuid']].values if name}
        elif self.level.lower() == 'pipeline':
            return {name: value for name, value in self._dsk.pipeline.list_knowledgepacks()[['Name', 'kp_uuid']].values if name}

        return {name: value for name, value in self._dsk.list_knowledgepacks()[['Name', 'kp_uuid']].values if name}

    @staticmethod
    def flatten(l):
        return [item for sublist in l for item in sublist]

    def get_model_list(self):
        return [Box([Label(value='Parent Model')], layout=category_item_layout)] + self.flatten(self.kb_dict['parent']) + \
            [Box([Label(value='Sub Model')], layout=category_item_layout)
             ] + self.flatten(self.kb_dict['sub'])

    def get_feature_file_list(self):
        ff = self._dsk.list_featurefiles(silent=True)
        if ff is not None:
            return list(ff['Name'].values)
        else:
            return []

    def get_platform_names(self):
        pf = {}
        for i in range(len(self.platforms)):
            pf['{} {}'.format(self.platforms.iloc[i]['Board'],
                              self.platforms.iloc[i]['Platform Version'])] = i

        return pf

    def update(self):

        self.test_data_widget.options = [None] + self.get_feature_file_list()

        self.kb_items[PARENT_INDEX].children[1].options = self.get_kp_dict()
        self.kb_items[SUBMODEL_INDEX].children[1].options = self.get_kp_dict()

        if self.kb_dict['parent']:
            for child in self.kb_dict['parent'][0][1:]:
                child.children[0].options = [
                    k for k, v in self.kb_description['sub'].items()]

    def create_widget(self):

        self.platform_widget = Dropdown(options=self.get_platform_names())
        self.target_os_widget = Dropdown()
        self.app_output_options_widget = SelectMultiple(
            options=['ble', 'led', 'serial'])
        self.download_type_widget = Dropdown(options=['Library', 'Binary'])
        self.imu_sample_widget = Dropdown(options=[100, 50, 25])
        self.debug_widget = Dropdown(options=[True, False])
        self.test_data_widget = Dropdown(options=[None])

        self.kb_items = [
            Box([Label(value='Model Builder'),
                 Button(icon='0', description="Generate Knowledgepack")], layout=form_item_layout),
            Box([Label(value='Select Hardware Platform', layout=label_value_layout), widgets.interactive(
                self.select_platform, platform_name=self.platform_widget)],  layout=form_item_layout),
            Box([Label(value='Select Target OS', layout=label_value_layout),
                 self.target_os_widget], layout=form_item_layout),
            Box([Label(value='Select Output', layout=label_value_layout),
                 self.app_output_options_widget], layout=form_item_layout),
            Box([Label(value='Download Type', layout=label_value_layout),
                 self.download_type_widget], layout=form_item_layout),
            Box([Label(value='IMU Sample Rate', layout=label_value_layout),
                 self.imu_sample_widget], layout=form_item_layout),
            Box([Label(value='Debug', layout=label_value_layout),
                 self.debug_widget], layout=form_item_layout),
            Box([Label(value='Test Data', layout=label_value_layout),
                 self.test_data_widget], layout=form_item_layout),
            Box([Label(value='Set Parent Model'),
                 Dropdown(options=[]),
                 Button(icon='1', description="Set")], layout=form_item_layout),
            Box([Label(value='Add Sub Model'),
                 Dropdown(options=[]),
                 Button(icon='2', description="Add")],
                layout=form_item_layout),
        ]

        self.kb_items[GENERATE_INDEX].children[1].on_click(
            self.generate_description)
        self.kb_items[PARENT_INDEX].children[2].on_click(self.set_parent_model)
        self.kb_items[SUBMODEL_INDEX].children[2].on_click(self.add_submodel)

        model_selector = Box(self.kb_items,
                             layout=pipeline_layout)
        display(model_selector)
        self.generate_pipeline()

    def generate_pipeline(self):
        if self._dsk is None:
            return

        self.update()
        try:
            self._pipeline.close()
        except:
            pass
        self.update()
        self._pipeline = (Box(self.get_model_list(), layout=pipeline_layout))
        display(self._pipeline)




from sensiml.method_calls.featurefilecall import FeatureFileCall
from sensiml.method_calls.capturefilecall import CaptureFileCall
from sensiml.method_calls.querycall import QueryCall
from sensiml.method_calls.functioncall import FunctionCall
from sensiml.method_calls.trainandvalidationcall import TrainAndValidationCall
from sensiml.method_calls.generatorcall import GeneratorCall
from sensiml.method_calls.generatorcallset import GeneratorCallSet
from sensiml.method_calls.selectorcall import SelectorCall
from sensiml.method_calls.selectorcallset import SelectorCallSet
from sensiml.method_calls.validationmethodcall import ValidationMethodCall
from sensiml.method_calls.classifiercall import ClassifierCall
from sensiml.method_calls.optimizercall import OptimizerCall
from sensiml.method_calls.trainingalgorithmcall import TrainingAlgorithmCall


__all__ = ['FunctionCall', 'QueryCall', 'FeatureFileCall', 'TrainingAlgorithmCall',  \
			'GeneratorCall', 'GeneratorCallSet', 'SelectorCall', 'SelectorCallSet', 'ValidationMethodCall', \
			'ClassifierCall', 'OptimizerCall', 'TrainingAlgorithmCall', 'CaptureFileCall']


import os
import json
import ctypes
from ctypes import CDLL
from pandas import DataFrame, Series
from numpy import ndarray
from wurlitzer import pipes


class ImportPathException(Exception):
    pass


class struct_pme_pattern(ctypes.Structure):
    __slots__ = ['influence', 'category', 'vector'
                 ]

    _fields_ = [('influence', ctypes.c_uint16),
                ('category', ctypes.c_uint16),
                ('vector', ctypes.POINTER(ctypes.c_uint8)),
                ]


class struct_pme_model_header(ctypes.Structure):
    __slots__ = ['number_patterns', 'pattern_length'
                 ]

    _fields_ = [('number_patterns', ctypes.c_uint16),
                ('pattern_length', ctypes.c_uint16),
                ]


def empty_function(*args, **kwargs):
    print("Not Supported by this Knowledge Pack.")
    return


class Model(object):
    def __init__(self, neuron_array, class_map, configuration, feature_summary=None):
        self._class_map = None
        self._feature_summary = None
        self._neuron_array = None
        self._configuration = None

        self.class_map = class_map
        self.neuron_array = neuron_array
        self.configuration = configuration

        if feature_summary:
            self.feature_summary = feature_summary
        else:
            self.feature_summary_from_neurons()

    @property
    def class_map(self):
        return self._class_map

    @class_map.setter
    def class_map(self, value):
        self._class_map = value

    @property
    def neuron_array(self):
        return self._neuron_array

    @neuron_array.setter
    def neuron_array(self, value):
        self._neuron_array = value

    @property
    def configuration(self):
        return self._configuration

    @configuration.setter
    def configuration(self, value):
        self._configuration = value

    @property
    def feature_summary(self):
        return self._feature_summary

    @feature_summary.setter
    def feature_summary(self, value):
        self._feature_summary = [{'Feature': str(x)} for x in value]

    def feature_summary_from_neurons(self):
        feature_vector = self.neuron_array[0]['Vector']
        self.feature_summary = [
            {'Feature': str(x)} for x in range(len(feature_vector))]


class SMLRunner(object):
    """
    This class provides a python interface to the knowledgpack libary generated by SensiML.
    After downloading the knowledgpack for your platform, you can instantiate this function
    by passing the path to libsensiml.so

        sml = SMLRunner(<path_to_libsensiml.so>)
        sml.init_model()

    Then, you can pythonically call all of the c functions in kb.h passing data like you
    normally would.  For example, if you have a dataframe df which has your sensor data
    stored in it,

        print(df.head(2))
        > AccX, AccY, AccZ
            9    10     11
            12   8      23

    you can call run_model and pass a single sample at a time to run the model as you would
    on the device.

        for index in range(len(df)):
            category = sml.run_model(df.iloc[index].values, 0, model_index)
            if category >= 0:
                print(category)
                sml.reset_model(0)

    """

    def __init__(self, path):
        self._run_type = None
        self._model_initialized = False
        if os.name == 'nt':
            print("SML Runner is not supported on Windows OS.")

        if not os.path.exists(os.path.join(path, "libsensiml.so")):
            print("ERROR: libsensiml.so as not found in {}.".format(path))
            raise ImportPathException(
                "ERROR: libsensiml.so as not found in {}".format(path))

        clf_lib = CDLL(os.path.join(path, "libsensiml.so"))
        self._model_init = clf_lib.kb_model_init
        self._model_init.argtypes = []

        self._run_model = clf_lib.kb_run_model
        self._run_model.argtypes = [ctypes.POINTER(
            ctypes.c_int16), ctypes.c_int, ctypes.c_int]
        self._run_model.restype = ctypes.c_int

        self._run_segment = clf_lib.kb_run_segment
        self._run_segment.argtypes = [ctypes.c_int]
        self._run_segment.restype = ctypes.c_int

        self._add_segment = clf_lib.kb_add_segment
        self._add_segment.argtypes = [ctypes.POINTER(
            ctypes.c_int16), ctypes.c_int, ctypes.c_int, ctypes.c_int]

        self._reset_model = clf_lib.kb_reset_model
        self._reset_model.argtypes = [ctypes.c_int]
        self._reset_model.restype = ctypes.c_int

        self._print_model_map = clf_lib.kb_print_model_map

        self._print_model_result = clf_lib.kb_print_model_result
        self._print_model_result.argtypes = [ctypes.c_int, ctypes.c_int]

        self._print_model = clf_lib.kb_print_model
        self._print_model.argtypes = [ctypes.c_int]
        self._print_model.restype = ctypes.c_int

        self._print_model_score = clf_lib.kb_print_model_score
        self._print_model_score.argtypes = [ctypes.c_int]
        self._print_model_score.restype = ctypes.c_int

        self._score_model = clf_lib.kb_score_model
        self._score_model.argtypes = [ctypes.c_int, ctypes.c_uint16]
        self._score_model.restype = ctypes.c_int

        self._retrain_model = clf_lib.kb_retrain_model
        self._retrain_model.argtypes = [ctypes.c_int]
        self._retrain_model.restype = ctypes.c_int

        self._add_last_pattern_to_model = clf_lib.kb_add_last_pattern_to_model
        self._add_last_pattern_to_model.argtypes = [
            ctypes.c_int, ctypes.c_uint16, ctypes.c_uint16]
        self._add_last_pattern_to_model.restype = ctypes.c_int

        self._add_custom_pattern_to_model = clf_lib.kb_add_custom_pattern_to_model
        self._add_custom_pattern_to_model.argtypes = [
            ctypes.c_int, ctypes.POINTER(ctypes.c_uint8), ctypes.c_uint16, ctypes.c_uint16]
        self._add_custom_pattern_to_model.restype = ctypes.c_int

        self._print_model_class_map = clf_lib.kb_print_model_class_map
        self._print_model_class_map.argtypes = [ctypes.c_int, ctypes.c_char_p]

        self._get_model_header = clf_lib.get_model_header
        self._get_model_header.argtypes = [
            ctypes.c_int, ctypes.POINTER(struct_pme_model_header)]
        self._get_model_header.restype = ctypes.c_int

        self._get_model_pattern = clf_lib.get_model_pattern
        self._get_model_pattern.argtypes = [
            ctypes.c_int, ctypes.c_int, ctypes.POINTER(struct_pme_pattern)]
        self._get_model_pattern.restype = ctypes.c_int

        self._flush_model = clf_lib.flush_model
        self._flush_model.argtypes = [ctypes.c_int]
        self._flush_model.restype = ctypes.c_int

        try:
            self._get_feature_vector = clf_lib.kb_get_feature_vector
            self._get_feature_vector.argtypes = [ctypes.c_int, ctypes.POINTER(
                ctypes.c_uint8), ctypes.POINTER(ctypes.c_uint8)]
        except:
            self._get_feature_vector = empty_function

    def _initialized(self):
        if os.name == 'nt':
            print("SML Runner is not supported on Windows OS.")
            return None

        if not self._model_initialized:
            print("Initialize the model before running this function")
            return None

        return True

    def _run_with(self, run_type):
        if self._run_type is None:
            self._run_type = run_type

        if self._run_type != run_type:
            print("This model has already been run with {}. You will need to restart your kernal to run in this mode.".format(
                self._run_type))
            return False

        return True

    def init_model(self):
        """
        This will initialize the parameters of all the models, it should be run once before
        running anything elese.
        """
        if not self._model_initialized:
            self._model_initialized = True
            self._model_init()

    def add_custom_pattern_to_model(self, model_index, feature_vector, category, aif):
        """
        updates the model by adding the a new pattern to the database
        with a new category and influence field

        Args:
           model_index(int): Index of the model to update
           feature_vector(list): List of uint8 values that will be added to the database
           category(uint16): category of the vector to add
           aif(uint16): weight function for the new pattern

         Returns:
             int: 0 if model does not support dynamic updates
                  -1 if model can not be updated anymore
                  1 if model was succesfully updated
         """

        if not self._initialized():
            return

        model_index_ctype = ctypes.c_int(model_index)
        category_ctype = ctypes.c_uint16(category)
        aif_ctype = ctypes.c_uint16(aif)

        feature_vector_array = (ctypes.c_uint8*len(feature_vector))()

        for index, value in enumerate(feature_vector):
            feature_vector_array[index] = ctypes.c_uint8(value)

        return self._add_custom_pattern_to_model(model_index_ctype, feature_vector_array, category_ctype, aif_ctype)

    def add_last_pattern_to_model(self, model_index, category, aif):
        """
        Updates the model by adding the a the last feature vector created to the database
        with a new category and influence field

        Args:
           model_index(int): Index of the model to update
           category(uint16): category of the vector to add
           aif(uint16): weight function for the new pattern

         Returns:
             int: 0 if model does not support dynamic updates
                  -1 if model can not be updated anymore
                  1 if model was succesfully updated
         """

        if not self._initialized():
            return

        model_index_ctype = ctypes.c_int(model_index)
        category_ctype = ctypes.c_uint16(category)
        aif_ctype = ctypes.c_uint16(aif)

        return self._add_last_pattern_to_model(model_index_ctype, category_ctype, aif_ctype)

    def score_model(self, model_index, category, silent=False):
        """
        Given the ground truth of the current feature vector, this will score the model
        internally

        Args:
           model_index(int): Index of the model to update
           category(uint16): category of the vector to add

        Returns:
           int: 0 if model does not support dynamic scoring
                -1 if model can not be scored
                1 if model was succesfully scored
        """
        if not self._initialized():
            return

        model_index_ctype = ctypes.c_int(model_index)
        category_ctype = ctypes.c_uint16(category)

        res = None
        with pipes() as (out, _):
            res = self._score_model(model_index_ctype, category_ctype)

        if silent:
            print(out.read())

        return res

    def retrain_model(self, model_index):
        """
        After a model has been scored, this will retraing the model according
        to the models supported reinforcment learning approach

        Args:
           model_index(int): Index of the model to update

        Returns:
           int: 0 if model does not support dynamic scoring
                1 if model was succesfully scored
        """
        if not self._initialized():
            return

        model_index_ctype = ctypes.c_int(model_index)

        ret = self._retrain_model(model_index_ctype)

        if ret == 0:
            print("Retraining is not supported/enabled for this model.")

        return ret

    def run_model(self, data_sample, model_index, debug_log=False):
        """
        This is the main entry point into the pipeline. It takes a single timepoint of data as an array. Adds that sample to the internal ring buffer
        checks for a segment, generates features if there is a segment,
        produces a classification and returns the result.

        Args:
            data_sample(array):  single timepoint of data as an array from all sensors
            nsensors(int): unused
            model_index(int): Index of the model to run
            debug_log(bool): Prints the debug output from the knowledge pack

         Returns:
             Classification results will be 0 if unkown through the classification numbers
                 you have. This function returns -1 when a segment hasn't yet been identified.

        """
        if not self._initialized():
            return

        if self._run_with('run_model') is False:
            return

        if isinstance(data_sample, DataFrame) or isinstance(data_sample, Series):
            data = data_sample.values
        elif isinstance(data_sample, ndarray):
            data = data_sample
        else:
            print("Input data Must be either dataframe or array.")
            return

        data_array = (ctypes.c_int16*data.shape[0])()
        nsensors_ctype = ctypes.c_int(0)
        model_index_ctype = ctypes.c_int(model_index)

        for index, value in enumerate(data):
            data_array[index] = ctypes.c_int16(value)

        with pipes() as (out, _):
            ret = self._run_model(
                data_array, nsensors_ctype, model_index_ctype)

        if debug_log:
            print(out.read())

        return ret

    def run_segment(self, data_segment, model_index, debug_log=False):
        """
        Add a segment of data to the model as input. Then runs the model on the current segment, skipping the data streaming steps.

        (Warning: using this call will flush the models internal ring buffer. Alternating between this and run_and_score model may produce incorrect results or crashes)

        Args:
            data(array):  Array of timseries data
            model_index(int): Index of the model to run
            debug_log(bool): Prints the debug output from the knowledge pack

         Returns:
             Classification results will be 0 if unkown through the classification numbers
                 you have. This function returns -1 when a segment hasn't yet been identified.

        """
        if not self._initialized():
            return

        if self._run_with('run_segment') is False:
            return

        if isinstance(data_segment, (DataFrame, Series)):
            data = data_segment.values
        elif isinstance(data_segment, ndarray):
            data = data_segment
        else:
            print("Input data Must be either dataframe or array.")
            return

        data_array = (ctypes.c_int16*(data.shape[0]*data.shape[1]))()
        length_ctype = ctypes.c_int(data.shape[0])
        nbuffs = ctypes.c_int(data.shape[1])
        model_index_ctype = ctypes.c_int(model_index)

        for col in range(data.shape[1]):
            for index, value in enumerate(data[:, col]):
                data_array[col*data.shape[0]+index] = ctypes.c_int16(value)

        self._add_segment(data_array, length_ctype, nbuffs, model_index_ctype)

        model_index_ctype = ctypes.c_int(model_index)

        with pipes() as (out, _):

            ret = self._run_segment(model_index_ctype)

        if debug_log:
            print(out.read())

        return ret

    def reset_model(self,  model_index):
        """
        Advances the model state to so that it is ready for a more input data. Use
        only after classification steps.

        (Note: this does not reset the model to a clean state, only init_model does this)


        Args:
           model_index(int): Index of the model to update
        """

        if not self._initialized():
            return

        model_index_ctype = ctypes.c_int(model_index)

        self._reset_model(model_index_ctype)

    def flush_model(self, model_index):
        """
        Deletes all of the patterns in the database.

        Args:
           model_index(int): Index of the model to update
        """

        if not self._initialized():
            return

        model_index_ctype = ctypes.c_int(model_index)

        self._flush_model(model_index_ctype)

    def get_model_map(self):
        """
            Print the model name to map relationship
        """

        with pipes() as (out, _):
            self._print_model_map()

        return json.loads(out.read())

    def get_model_result(self, model_index, result):
        """

        """

        if not self._initialized():
            return

        model_index_ctype = ctypes.c_int(model_index)
        result_ctype = ctypes.c_int(result)

        with pipes() as (out, _):
            self._print_model_result(model_index_ctype, result_ctype)

        return json.loads(out.read())

    def get_model(self, model_index):
        """
        Prints the model weights and info for current model if supported

        Args:
            model index - number of axes in the data stream

        """

        if not self._initialized():
            return

        model_header = self.get_model_header(model_index)
        model = []


        for index in range(model_header.number_patterns):
            tmp_dict = {}
            pattern = self.get_model_pattern(model_index, index)
            tmp_dict['Category'] = pattern.category
            tmp_dict['Vector'] = [pattern.vector[i] for i in range(model_header.pattern_length)]
            tmp_dict['AIF'] = pattern.influence
            tmp_dict['Identifier'] = index
            model.append(tmp_dict)

        return DataFrame(model)

    def get_model_score(self, model_index):

        if not self._initialized():
            return

        model_index_ctype = ctypes.c_int(model_index)

        with pipes() as (out, _):
            self._print_model_score(model_index_ctype)

        model_scores = []
        for line in out.read().split('\n')[:-1]:
            model_scores.append(json.loads(line))

        cats = [str(x)
                for x in range(1, len(DataFrame(model_scores).columns)-1)]

        return DataFrame(model_scores)[["ID", 'ERR']+cats].style.apply(_color, axis=1)

    def get_class_map(self, model_index):

        if not self._initialized():
            return

        model_index_ctype = ctypes.c_int(model_index)

        with pipes() as (out, _):
            self._print_model_class_map(model_index_ctype, None)

        return json.loads(out.read())

    def get_model_header(self, model_index):

        if not self._initialized():
            return

        model_index_ctype = ctypes.c_int(model_index)
        pme_model_header = struct_pme_model_header()

        with pipes() as (out, _):
            self._get_model_header(model_index_ctype, pme_model_header)

        return pme_model_header

    def get_model_pattern(self, model_index, pattern_index):

        if not self._initialized():
            return

        model_index_ctype = ctypes.c_int(model_index)
        pattern_index_ctype = ctypes.c_int(pattern_index)
        pme_pattern = struct_pme_pattern()

        with pipes() as (out, _):
            self._get_model_pattern(
                model_index_ctype, pattern_index_ctype, pme_pattern)

        return pme_pattern

    def get_feature_vector(self, model_index):

        if not self._initialized():
            return

        model_index_ctype = ctypes.c_int(model_index)
        feature_vector_ctype = (ctypes.c_ubyte*128)()
        feature_length_ctype = ctypes.c_ubyte(0)

        self._get_feature_vector(
            model_index_ctype, feature_vector_ctype, ctypes.byref(feature_length_ctype))

        return list(feature_vector_ctype[:feature_length_ctype.value])

    def knowledgepack(self, model_index, distance_mode=0, feature_list=None):
        """
        Turn into a knowledgepack Model Object (Only works for PME)
        """
        self.init_model()
        neuron_array = self.get_model(model_index).to_dict('records')
        class_map = self.get_class_map(model_index)

        return Model(neuron_array, class_map, {'distance_mode': distance_mode}, feature_summary=feature_list)

    def run_and_score_model(self, data_sample, category, model_index=0):
        """
        Will Run ans core a model based on passing a single sample at a time.

        """

        ret = self.run_model(data_sample, model_index)

        if ret is None:
            return

        if ret >= 0:
            print('Class: {}'.format(ret))
            if ret > 0:  # only retrain if greater than 0
                score_ret = self.score_model(model_index, category)
                if score_ret == 0:
                    print("Retraining is not supported for this model.")
            self.reset_model(model_index)

            return self.get_model_score(model_index)

        return -1

    def run_and_score_model_on_segment(self, segment, category, model_index=0):
        """
        will run and score a model based on a segment of data.

        (Warning: using this call will flush the models internal ring buffer. Alternating between this and run_and_score model may produce incorrect results or crashes)

        """

        ret = self.run_segment(segment, model_index)

        if ret is None:
            return

        if ret >= 0:
            print('Class: {}'.format(ret))
            if ret > 0:  # only retrain if greater than 0
                self.score_model(model_index, category)
            self.reset_model(model_index)

        return self.get_model_score(model_index)


def _color(val):
    color = 'white'
    if val.ERR > 0:
        color = '#90ee90'
    elif val.ERR < 0:
        color = 'red'
    return ['background-color: %s' % color]*len(val)



import sys
import os
from setuptools import setup, find_packages

__version__ = "2019.3.1"

# 'setup.py publish' shortcut.
if sys.argv[-1] == 'publish':
    os.system('python setup.py sdist bdist_wheel')
    os.system('twine upload dist/*')
    sys.exit()

# 'setup.py test' shortcut.
# !pip install --index-url https://test.pypi.org/simple/ sensiml -U
if sys.argv[-1] == 'test':
    os.system('python setup.py sdist bdist_wheel')
    os.system('twine upload --repository-url https://test.pypi.org/legacy/ dist/*')
    sys.exit()

setup(
    name='SensiML',
    description='SensiML Analytic Suite Python client',
    version=__version__,
    author='SensiML',
    author_email='support@sensiml.com',
    license='Proprietary',
    classifiers=[
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    packages=find_packages(exclude=['*test*', ]),
    package_data={
        'sensiml.datasets': ['*.csv'],
        'sensiml.widgets': ['*.pem'],
        'sensiml.image': ['*.png'],
    },
    include_package_data=True,
    long_description=open('README.md').read(),
    install_requires=[
        'cookiejar',
        'requests>=2.14.2',
        'requests-oauthlib>=0.7.0',
        'appdirs>=1.4.3',
        'semantic_version>=2.6.0',
        'jupyter>=1.0.0',
        'numpy>=1.13',
        'pandas>=0.20.3',
        'matplotlib>=2.0.0',
        'nrfutil>=3.3.2,<=5.0.0',
        'qgrid>=1.0.2',
        'prompt-toolkit>=2.0.5; python_version>="3"',
        'prompt-toolkit<=1.0.4; python_version<"3"',
        'jupyter-console>=6.0.0; python_version>="3"',
        'notebook==5.7.5; python_version>="3"',
        'jupyter-console<=5.2.0; python_version<"3"',
        'ipython>=7.0.1; python_version>="3"',
        'ipython<=5.8.0; python_version<"3"',
        'bqplot',
        'seaborn',
        'wurlitzer',
        'jupyter-contrib-nbextensions',
        'pyserial',
    ],
)
